/**
 * @file c64net.c
 * Transfer routines for the 2-bit/4-bit PET/VIC-20/C64/C128 to PC cable
 * used by the 64NET software
 * @author Marko Mkel (marko.makela@nic.funet.fi)
 */

/*
 * Copyright  1994-1996 Marko Mkel and Olaf Seibert
 * Original Linux and Commodore 64/128/Vic-20 version by Marko Mkel
 * Ported to the PET and the Amiga series by Olaf Seibert
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef COMM_PC
# include "info.h"
# include "c64net.h"
# include "pcpar.h"

/** Open the communication channel
 * @param dev		name of the communication device
 * @param hostinfo	(output) the computer model information
 * @return		zero on success, nonzero on failure
 */
int
c64net_init (const char* dev, struct hostinfo* hostinfo)
{
  unsigned char detectbuf[5];
  if (!open_port (dev))
    return 1;
  outb (0, baseaddr);
  if (c64net_write ("", 1))
    return 1;
  if (c64net_read (detectbuf, sizeof detectbuf))
    return 2;
  hostinfo->host = detectbuf[0];
  hostinfo->driver = (unsigned) detectbuf[1] | (unsigned) detectbuf[2] << 8;
  hostinfo->basic = (unsigned) detectbuf[3] | (unsigned) detectbuf[4] << 8;
  return 0;
}

/** Close the communication channel */
void
c64net_close (void)
{
  outb (0, baseaddr);
}

/** Send data
 * @param buf		the data to be sent
 * @param len		length of the data in bytes
 * @return		zero on success, nonzero on failure
 */
int
c64net_write (const void* buf, unsigned len)
{
# ifdef __GNUC__
  __asm ("movw baseaddr,%%dx\n\t"
	 "movl %1,%%ecx\n\t"
	 "movl %0,%%ebx\n"
	 "sloop:\t"
	 "movb (%%ebx),%%al\n\t"
	 "orb $8,%%al\n\t"
	 "outb %%al,%%dx\n\t"
	 "incw %%dx\n"
	 "wait1:\t"
	 "inb %%dx,%%al\n\t"
	 "testb $32,%%al\n\t"
	 "jne wait1\n\t"
	 "movb %0,%%al\n\t"
	 "salb $4,%%al\n\t"
	 "decw %%dx\n\t"
	 "outb %%al,%%dx\n\t"
	 "incw %%dx\n"
	 "wait2:\t"
	 "inb %%dx,%%al\n\t"
	 "testb $32,%%al\n\t"
	 "je wait2\n\t"
	 "decw %%dx\n\t"
	 "inc %%ebx\n\t"
	 "loop sloop"
	 : /* no outputs */
	 : "g" (buf), "g" (len)
	 : "al", "ebx", "ecx", "dx");
# elif defined _MSC_VER || defined __BCC__
#  ifdef _MSC_VER
  __asm {
#  endif
#  ifdef __BCC__
#  asm
#  endif
    push ds
    mov dx,baseaddr
    mov cx,len
    lds bx,buf
sloop:
    mov al,[bx]
    or al,8
    out dx,al

    inc dx
wait1:
    in al,dx
    test al,32
    je nwait1
    in al,0x60  ; is the esc key pressed?
    dec al
    jnz wait1
    jz abort

nwait1:
    mov al,[bx]
#  if defined _M_I286 || defined _M_I386
    shl al,4
#  else
    shl al,1
    shl al,1
    shl al,1
    shl al,1
#  endif
    dec dx
    out dx,al

    inc dx
wait2:
    in al,dx
    test al,32
    jne nwait2
    in al,0x60  ; is the esc key pressed?
    dec al
    jnz wait1
    jz abort
nwait2:
    dec dx

    inc bx
    loop sloop
abort:
    pop ds
#  ifdef __MSC_VER
  }
#  endif
#  ifdef __BCC__
#   endasm
#  endif

  if (inb (0x60) == 1) /* ESC pressed during the transfer */
    exit (-1);
# else
  register const unsigned char* buffer = buf;

  while (len--) {
    outb (*buffer | 8, baseaddr);
    while(0x20 & inb (stataddr));
    outb (*buffer++ << 4, baseaddr);
    while (!(0x20 & inb (stataddr)));
  }
# endif
  return 0;
}

/** Receive data
 * @param buf		the data to be received
 * @param len		length of the data in bytes
 * @return		zero on success, nonzero on failure
 */
int
c64net_read (void* buf, unsigned len)
{
# ifdef __GNUC__
  __asm ("movw baseaddr,%%dx\n\t"
	 "movl %1,%%ecx\n\t"
	 "movl %0,%%ebx\n"
	 "rloop:\t"
	 "movb $0xf0,%%al\n\t"
	 "outb %%al,%%dx\n\t"
	 "incw %%dx\n"
	 "input0:\t"
	 "inb %%dx,%%al\n\t"
	 "testb $32,%%al\n\t"
	 "jne input0\n\t"
	 "decw %%dx\n\t"
	 "xorb %%al,%%al\n\t"
	 "outb %%al,%%dx\n\t"
	 "incw %%dx\n"
	 "input1:\t"
	 "inb %%dx,%%al\n\t"
	 "testb $32,%%al\n\t"
	 "je input1\n\t"
	 "decw %%dx\n\t"
	 "andb $0xc0,%%al\n\t"
	 "shrb $6,%%al\n\t"
	 "mov %%al,(%%ebx)\n\t"
	 "movb $8,%%al\n\t"
	 "outb %%al,%%dx\n\t"
	 "incw %%dx\n"
	 "input2:\t"
	 "inb %%dx,%%al\n\t"
	 "testb $32,%%al\n\t"
	 "jne input2\n\t"
	 "decw %%dx\n\t"
	 "andb $0xc0,%%al\n\t"
	 "shrb $4,%%al\n\t"
	 "orb %%al,(%%ebx)\n\t"
	 "xorb %%al,%%al\n\t"
	 "outb %%al,%%dx\n\t"
	 "incw %%dx\n"
	 "input3:\t"
	 "inb %%dx,%%al\n\t"
	 "testb $32,%%al\n\t"
	 "je input3\n\t"
	 "decw %%dx\n\t"
	 "andb $0xc0,%%al\n\t"
	 "shrb $2,%%al\n\t"
	 "orb %%al,(%%ebx)\n\t"
	 "movb $8,%%al\n\t"
	 "outb %%al,%%dx\n\t"
	 "incw %%dx\n"
	 "input4:\t"
	 "inb %%dx,%%al\n\t"
	 "testb $32,%%al\n\t"
	 "jne input4\n\t"
	 "decw %%dx\n\t"
	 "andb $0xc0,%%al\n\t"
	 "xorb $0xaa,%%al\n\t"
	 "xorb %%al,(%%ebx)\n\t"
	 "xorb %%al,%%al\n\t"
	 "outb %%al,%%dx\n\t"
	 "incw %%dx\n"
	 "input5:\t"
	 "inb %%dx,%%al\n\t"
	 "testb $32,%%al\n\t"
	 "je input5\n\t"
	 "decw %%dx\n\t"
	 "inc %%ebx\n\t"
	 "loop rloop"
	 : /* no outputs */
	 : "g" (buf), "g" (len)
	 : "ax", "ebx", "ecx", "dx");
# elif defined _MSC_VER || defined __BCC__
#  ifdef _MSC_VER
  __asm {
#  endif
#  ifdef __BCC__
#  asm
#  endif
    push ds
    mov dx,baseaddr
    mov cx,len
    lds bx,buf

rloop:
    mov al,0xf0
    out dx,al
    inc dx
input0:
    in al,dx
    test al,32
    je ninput0
    in al,0x60  ; is the esc key pressed?
    dec al
    jnz input0
    jmp abort

ninput0:
    dec dx

    xor al,al
    out dx,al
    inc dx

input1:
    in al,dx
    test al,32
    jne ninput1
    in al,0x60  ; is the esc key pressed?
    dec al
    jnz input1
    jz abort
ninput1:
    dec dx
    and al,0xc0
#  if defined _M_I286 || defined _M_I386
    shr al,6
#  else
    shr al,1
    shr al,1
    shr al,1
    shr al,1
    shr al,1
    shr al,1
#  endif
    mov [bx],al

    mov al,8
    out dx,al
    inc dx

input2:
    in al,dx
    test al,32
    je ninput2
    in al,0x60  ; is the esc key pressed?
    dec al
    jnz input2
    jz abort
ninput2:
    dec dx
    and al,0xc0
#  if defined _M_I286 || defined _M_I386
    shr al,4
#  else
    shr al,1
    shr al,1
    shr al,1
    shr al,1
#  endif
    or [bx],al

    xor al,al
    out dx,al
    inc dx

input3:
    in al,dx
    test al,32
    jne ninput3
    in al,0x60  ; is the esc key pressed?
    dec al
    jnz input3
    jz abort
ninput3:
    dec dx
    and al,0xc0
#  if defined _M_I286 || defined _M_I386
    shr al,2
#  else
    shr al,1
    shr al,1
#  endif
    or [bx],al

    mov al,8
    out dx,al
    inc dx

input4:
    in al,dx
    test al,32
    je ninput4
    in al,0x60  ; is the esc key pressed?
    dec al
    jnz input4
    jz abort
ninput4:
    dec dx
    and al,0xc0
    xor al,0xaa
    xor [bx],al

    xor al,al
    out dx,al
    inc dx

input5:
    in al,dx
    test al,32
    jne ninput5
    in al,0x60  ; is the esc key pressed?
    dec al
    jnz input5
    jz abort
rlop:
    jmp rloop	; otherwise the branch would be out of range
ninput5:
    dec dx

    inc bx
    loop rlop	; otherwise the branch would be out of range
abort:
    pop ds
#  ifdef __MSC_VER
  }
#  endif
#  ifdef __BCC__
#   endasm
#  endif

  if (inb (0x60) == 1) /* ESC pressed during the transfer */
    exit (-1);
# else
  register unsigned char* buffer = buf;

  while (len--) {
    register unsigned char data;
    outb (0xf0, baseaddr);
    while(0x20 & inb (stataddr));
    outb (0, baseaddr);

    while(!(0x20 & inb (stataddr)));
    data = (0xc0 & inb (stataddr)) >> 6;
    outb (8, baseaddr);

    while (0x20 & inb (stataddr));
    data |= (0xc0 & inb (stataddr)) >> 4;
    outb (0, baseaddr);

    while(!(0x20 & inb (stataddr)));
    data |= (0xc0 & inb (stataddr)) >> 2;
    outb (8, baseaddr);

    while (0x20 & inb (stataddr));
    data |= 0xc0 & inb (stataddr);
    outb (0, baseaddr);

    *buffer++ = data ^ 0xaa;

    while (!(0x20 & inb (stataddr)));
  }
# endif
  return 0;
}

#endif /* COMM_PC */
