/**
 * @file kontros.c
 * Transfer routines for the 8-bit/8-bit C64/C128/Vic-20 to PC cable
 * designed by Frank Kontros (frank@kontros.uzhgorod.ua)
 * @author Marko Mkel (marko.makela@nic.funet.fi)
 */

/*
 * Copyright  1994-1997 Marko Mkel and Olaf Seibert
 * Copyright  2001 Marko Mkel
 * Original Linux and Commodore 64/128/Vic-20 version by Marko Mkel
 * Ported to the PET and the Amiga series by Olaf Seibert
 * Restructured by Marko Mkel
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef COMM_PC
# include "info.h"
# include "kontros.h"
# include "pcpar.h"

/** Open the communication channel
 * @param dev		name of the communication device
 * @param hostinfo	(output) the computer model information
 * @return		zero on success, nonzero on failure
 */
int
kontros_init (const char* dev, struct hostinfo* hostinfo)
{
  unsigned char detectbuf[5];
  if (!open_port (dev))
    return 1;
  outb (0, ctrladdr); /* A/B=1, STROBE=1 */
  if (kontros_write ("", 1))
    return 1;
  kontros_sr ();
  if (kontros_read (detectbuf, sizeof detectbuf))
    return 2;
  kontros_rs ();
  hostinfo->host = detectbuf[0];
  hostinfo->driver = (unsigned) detectbuf[1] | (unsigned) detectbuf[2] << 8;
  hostinfo->basic = (unsigned) detectbuf[3] | (unsigned) detectbuf[4] << 8;
  return 0;
}

/** Close the communication channel */
void
kontros_close (void)
{
  outb (0, ctrladdr);
}

/** Switch the data direction from receive to send */
void
kontros_rs (void)
{
  while (!(inb(stataddr) & 0x80));
}

/** Switch the data direction from send to receive */
void
kontros_sr (void)
{
  outb (1, ctrladdr); /* -STROBE=0 */
  outb (0, ctrladdr); /* -STROBE=1 */
}

/** Send data
 * @param buf		the data to be sent
 * @param len		length of the data in bytes
 * @return		zero on success, nonzero on failure
 */
int
kontros_write (const void* buf, unsigned len)
{
# ifdef __GNUC__
  __asm ("movw baseaddr,%%dx\n\t"
	 "movl %1,%%ecx\n\t"
	 "movl %0,%%ebx\n"
	 "sloop:\t"
	 "movb (%%ebx),%%al\n\t"
	 "outb %%al,%%dx\n\t"
	 "addw $2,%%dx\n\t"
	 "movb $9,%%al\n\t"
	 "outb %%al,%%dx\n\t"
	 "decb %%al\n\t"
	 "outb %%al,%%dx\n\t"
	 "decw %%dx\n"
	 "owait:\t"
	 "inb %%dx,%%al\n\t"
	 "rolb $1,%%al\n\t"
	 "jnc owait\n\t"
	 "decw %%dx\n\t"
	 "inc %%ebx\n\t"
	 "loop sloop"
	 : /* no outputs */
	 : "g" (buf), "g" (len)
	 : "ax", "ebx", "ecx", "dx");
# elif defined _MSC_VER || defined __BCC__
#  ifdef _MSC_VER
  __asm {
#  endif
#  ifdef __BCC__
#  asm
#  endif
    mov dx,baseaddr
    mov cx,len

    push ds
    lds bx,buf

sloop:
    mov al,[bx]
    out dx,al
    add dx,2
    mov al,9
    out dx,al
    dec al
    out dx,al
    dec dx

owait:
    in al,dx
    rol al,1
    jc nowait
    in al,0x60 /* esc pressed? */
    dec al
    jne owait
    je abort
nowait:
    decw dx

    inc bx
    loop sloop

abort:
    pop ds
#  ifdef __MSC_VER
  }
#  endif
#  ifdef __BCC__
#   endasm
#  endif

  if (inb (0x60) == 1) /* ESC pressed during the transfer */
    exit (-1);
# else
  register const unsigned char* buffer = buf;

  while (len--) {
    register unsigned char byte = *buffer++;
    outb (byte, baseaddr);
    outb (9, ctrladdr); /* STROBE=0, SEL=0 (change port direction) */
    outb (8, ctrladdr); /* STROBE=1 */
    while (!(0x80 & inb (stataddr)));
  }
# endif
  return 0;
}

/** Receive data
 * @param buf		the data to be received
 * @param len		length of the data in bytes
 * @return		zero on success, nonzero on failure
 */
int
kontros_read (void* buf, unsigned len)
{
# ifdef __GNUC__
  __asm ("movw baseaddr,%%dx\n\t"
	 "movl %1,%%ecx\n\t"
	 "movl %0,%%ebx\n\t"
	 "incw %%dx\n"
	 "rloop:\t"
	 "movb $2,%%ah\n"
	 "iwait:\t"
	 "inb %%dx,%%al\n\t"
	 "rolb $1,%%al\n\t"
	 "jnc iwait\n\t"
	 "incw %%dx\n\t"
	 "xchgb %%ah,%%al\n\t"
	 "outb %%al,%%dx\n\t"
	 "decw %%dx\n\t"
	 "inb %%dx,%%al\n\t"
	 "rorb $3,%%al\n\t"
	 "orb $16,%%al\n\t"
	 "andb %%ah,%%al\n\t"
	 "movb %%al,(%%ebx)\n\t"
	 "incw %%dx\n\t"
	 "movb $1,%%al\n\t"
	 "outb %%al,%%dx\n\t"
	 "xorb %%al,%%al\n\t"
	 "outb %%al,%%dx\n\t"
	 "decw %%dx\n\t"
	 "inc %%ebx\n\t"
	 "loop rloop"
	 : /* no outputs */
	 : "g" (buf), "g" (len)
	 : "ax", "ebx", "ecx", "dx");
# elif defined _MSC_VER || defined __BCC__
#  ifdef _MSC_VER
  __asm {
#  endif
#  ifdef __BCC__
#  asm
#  endif
    mov dx,baseaddr
    mov cx,len

    push ds
    lds bx,buf

    inc dx
rloop:
    mov ah,2
iwait:
    in al,dx
    rol al,1
    jc niwait
    in al,0x60 /* esc pressed? */
    dec al
    jne iwait
    je abort

niwait:
    inc dx
    xchg al,ah
    out dx,al
    dec dx
    in al,dx
    ror al,3
    or al,16
    and al,ah
    mov [bx],al
    inc dx
    mov al,1
    out dx,al
    xor al,al
    out dx,al
    dec dx
    inc bx
    loop rloop

abort:
    pop ds
#  ifdef __MSC_VER
  }
#  endif
#  ifdef __BCC__
#   endasm
#  endif

  if (inb (0x60) == 1) /* ESC pressed during the transfer */
    exit (-1);
# else
  register unsigned char data1, data;
  register unsigned char* buffer = buf;

  while (len--) {
    /* get 5 bits (higher 4 bits) and wait for handshaking */
    while (!(0x80 & (data1 = inb (stataddr)))); /* wait for -BUSY to change */

    outb (2, ctrladdr); /* A/B=0 (lower bits from LS257) */
    data = (inb (stataddr) >> 3) & 0x0f; /* get lower 4 bits */
    data |= (data1 << 1) & 0xf0; /* combine the bits */

    outb (1, ctrladdr); /* -STROBE=0 */
    outb (0, ctrladdr); /* -STROBE=1 */

    *buffer++ = data;
  }
# endif
  return 0;
}

#endif /* COMM_PC */
