/**
 * @file pc64.c
 * Transfer routines for the 4-bit PET/C64/C128/Vic-20 to PC cable
 * designed by Wolfgang Lorenz
 * @author Marko Mkel (marko.makela@nic.funet.fi)
 */

/*
 * Copyright  1994-1996 Marko Mkel and Olaf Seibert
 * Copyright  2001 Marko Mkel
 * Original Linux and Commodore 64/128/Vic-20 version by Marko Mkel
 * Ported to the PET and the Amiga series by Olaf Seibert
 * Restructured by Marko Mkel
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef COMM_PC
# include "info.h"
# include "pc64.h"
# include "pcpar.h"

/** Open the communication channel
 * @param dev		name of the communication device
 * @param hostinfo	(output) the computer model information
 * @return		zero on success, nonzero on failure
 */
int
pc64_init (const char* dev, struct hostinfo* hostinfo)
{
  unsigned char detectbuf[5];
  if (!open_port (dev))
    return 1;
  outb (128, baseaddr); /* set the -FLAG signal to high */
  if (pc64_write ("", 1))
    return 1;
  if (pc64_read (detectbuf, sizeof detectbuf))
    return 2;
  hostinfo->host = detectbuf[0];
  hostinfo->driver = (unsigned) detectbuf[1] | (unsigned) detectbuf[2] << 8;
  hostinfo->basic = (unsigned) detectbuf[3] | (unsigned) detectbuf[4] << 8;
  return 0;
}
/** Close the communication channel */
void
pc64_close (void)
{
}

/** Send data
 * @param buf		the data to be sent
 * @param len		length of the data in bytes
 * @return		zero on success, nonzero on failure
 */
int
pc64_write (const void* buf, unsigned len)
{
# ifdef __GNUC__
  __asm ("movw baseaddr,%%dx\n\t"
	 "movl %1,%%ecx\n\t"
	 "movl %0,%%ebx\n"
	 "sloop:\t"
	 "movb (%%ebx),%%al\n\t"
	 "andb $0x0f,%%al\n\t"
	 "outb %%al,%%dx\n\t"
	 "incw %%dx\n"
	 "ewait:\t"
	 "inb %%dx,%%al\n\t"
	 "testb $0x80,%%al\n\t"
	 "jz ewait\n\t"
	 "decw %%dx\n\t"
	 "movb $0x80,%%al\n\t"
	 "outb %%al,%%dx\n\t"
	 "movb (%%ebx),%%al\n\t"
	 "shrb $4,%%al\n\t"
	 "outb %%al,%%dx\n\t"
	 "incw %%dx\n"
	 "zwait:\t"
	 "inb %%dx,%%al\n\t"
	 "testb $0x80,%%al\n\t"
	 "jnz zwait\n\t"
	 "decw %%dx\n\t"
	 "movb $0x80,%%al\n\t"
	 "outb %%al,%%dx\n\t"
	 "inc %%ebx\n\t"
	 "loop sloop"
	 : /* no outputs */
	 : "g" (buf), "g" (len)
	 : "ax", "ebx", "ecx", "dx");
# elif defined _MSC_VER || defined __BCC__
#  ifdef _MSC_VER
  __asm {
#  endif
#  ifdef __BCC__
#   asm
#  endif
    push ds
    mov dx,baseaddr
    mov cx,len
    lds bx,buf
sloop:
    mov al,[bx]
    and al,0x0f
    out dx,al
    inc dx

ewait:
    in al,dx
    test al,0x80
    jnz newait
    in al,0x60  ; is the esc key pressed?
    dec al
    jnz ewait
    jz abort

newait:
    dec dx
    mov al,0x80
    out dx,al
    mov al,[bx]
#  if defined _M_I286 || defined _M_I386
    shr al,4
#  else
    shr al,1
    shr al,1
    shr al,1
    shr al,1
#  endif
    out dx,al
    inc dx

zwait:
    in al,dx
    test al,0x80
    jz nzwait
    in al,0x60  ; is the esc key pressed?
    dec al
    jnz zwait
    jz abort

nzwait:
    dec dx
    mov al,0x80
    out dx,al

    inc bx
    loop sloop
abort:
    pop ds
#  ifdef _MSC_VER
  }
#  endif
#  ifdef __BCC__
#   endasm
#  endif

  if (inb (0x60) == 1) /* ESC pressed during the transfer */
    exit (-1);
# else
  register const unsigned char* buffer = buf;

  while (len--) {
    register unsigned char byte = *buffer++;

    outb (byte & 0x0f, baseaddr);
    while (!(128 & inb (stataddr)));
    outb (128, baseaddr);
    outb (byte >> 4, baseaddr);
    while (128 & inb (stataddr));
    outb (128, baseaddr);
  }
# endif
  return 0;
}

/** Receive data
 * @param buf		the data to be received
 * @param len		length of the data in bytes
 * @return		zero on success, nonzero on failure
 */
int
pc64_read (void* buf, unsigned len)
{
# ifdef __GNUC__
  __asm ("movw baseaddr,%%dx\n\t"
	 "movl %1,%%ecx\n\t"
	 "movl %0,%%ebx\n"
	 "rloop:\t"
	 "xorb %%al,%%al\n\t"
	 "outb %%al,%%dx\n\t"
	 "incw %%dx\n"
	 "dwait:\t"
	 "inb %%dx,%%al\n\t"
	 "testb $0x80,%%al\n\t"
	 "jz dwait\n\t"
	 "andb $0x78,%%al\n\t"
	 "shrb $3,%%al\n\t"
	 "movb %%al,(%%ebx)\n\t"
	 "decw %%dx\n\t"
	 "movb $0x80,%%al\n\t"
	 "outb %%al,%%dx\n\t"
	 "xorb %%al,%%al\n\t"
	 "outb %%al,%%dx\n\t"
	 "incw %%dx\n"
	 "vwait:\t"
	 "inb %%dx,%%al\n\t"
	 "testb $0x80,%%al\n\t"
	 "jnz vwait\n\t"
	 "andb $0x78,%%al\n\t"
	 "shlb $1,%%al\n\t"
	 "orb %%al,(%%ebx)\n\t"
	 "decw %%dx\n\t"
	 "movb $0x80,%%al\n\t"
	 "outb %%al,%%dx\n\t"
	 "inc %%ebx\n\t"
	 "loop rloop"
	 : /* no outputs */
	 : "g" (buf), "g" (len)
	 : "ax", "ebx", "ecx", "dx");
# elif defined _MSC_VER || defined __BCC__
#  ifdef _MSC_VER
  __asm {
#  endif
#  ifdef __BCC__
#  asm
#  endif
    push ds
    mov dx,baseaddr
    mov cx,len
    lds bx,buf

rloop:
    xor al,al
    out dx,al
    inc dx

dwait:
    in al,dx
    test al,0x80
    jnz ndwait
    in al,0x60  ; is the esc key pressed?
    dec al
    jnz dwait
    jz abort

ndwait:
    and al,0x78
#  if defined _M_I286 || defined _M_I386
    shr al,3
#  else
    shr al,1
    shr al,1
    shr al,1
#  endif
    mov [bx],al
    dec dx
    mov al,0x80
    out dx,al
    xor al,al
    out dx,al
    inc dx

vwait:
    in al,dx
    test al,0x80
    jz nvwait
    in al,0x60  ; is the esc key pressed?
    dec al
    jnz vwait
    jz abort

nvwait:
    and al,0x78
    shl al,1
    or [bx],al
    dec dx
    mov al,0x80
    out dx,al

    inc bx
    loop rloop

abort:
    pop ds
#  ifdef _MSC_VER
  }
#  endif
#  ifdef __BCC__
#   endasm
#  endif

  if (inb (0x60) == 1) /* ESC pressed during the transfer */
    exit (-1);
# else
  register unsigned char data;
  register unsigned char* buffer = buf;

  while (len--) {
    outb (0, baseaddr);
    while (!(128 & inb (stataddr)));
    outb (128, baseaddr); 
    data = (inb (stataddr) & 0x78) >> 3;
    outb (0, baseaddr);
    while (128 & inb (stataddr));
    *buffer++ = data | ((inb (stataddr) & 0x78) << 1);  
    outb (128, baseaddr);
  }
# endif
  return 0;
}

#endif /* COMM_PC */
