/**
 * @file prlink.c
 * IBM PC transfer routines for the 4-bit/8-bit PET/C64/C128/Vic-20
 * to Centronics cable designed by Marko Mkel
 * @author Marko Mkel (marko.makela@nic.funet.fi)
 */

/*
 * Copyright  1994-1996 Marko Mkel and Olaf Seibert
 * Copyright  2001 Marko Mkel
 * Original Linux and Commodore 64/128/Vic-20 version by Marko Mkel
 * Ported to the PET and the Amiga series by Olaf Seibert
 * Restructured by Marko Mkel
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef COMM_PC
# include "info.h"
# include "prlink.h"
# include "pcpar.h"

/** current state of the acknowledgement line (BUSY) */
static unsigned ack ASM(ack);

/** Open the communication channel
 * @param dev		name of the communication device
 * @param hostinfo	(output) the computer model information
 * @return		zero on success, nonzero on failure
 */
int
prlink_init (const char* dev, struct hostinfo* hostinfo)
{
  unsigned char detectbuf[5];
  if (!open_port (dev))
    return 1;

  outb (8, baseaddr); /* set the -FLAG signal to high */
  stataddr = baseaddr + 1;
  ctrladdr = baseaddr + 2;
  ack = 0x80 & inb (stataddr); /* get the current state of the BUSY line */

  if (prlink_write ("", 1))
    return 1;
  if (prlink_read (detectbuf, sizeof detectbuf))
    return 2;
  hostinfo->host = detectbuf[0];
  hostinfo->driver = (unsigned) detectbuf[1] | (unsigned) detectbuf[2] << 8;
  hostinfo->basic = (unsigned) detectbuf[3] | (unsigned) detectbuf[4] << 8;
  return 0;
}
/** Close the communication channel */
void
prlink_close (void)
{
}

/** Send data
 * @param buf		the data to be sent
 * @param len		length of the data in bytes
 * @return		zero on success, nonzero on failure
 */
int
prlink_write (const void* buf, unsigned len)
{
# ifdef __GNUC__
  __asm ("movw baseaddr,%%dx\n\t"
	 "movl %1,%%ecx\n\t"
	 "movl %0,%%ebx\n\t"
	 "movb ack,%%ah\n"
	 "sloop:\t"
	 "movb (%%ebx),%%al\n\t"
	 "xorb $0xb,%%al\n\t"
	 "addw $2,%%dx\n\t"
	 "outb %%al,%%dx\n\t"
	 "subw $2,%%dx\n\t"
	 "andb $0xf7,%%al\n\t"
	 "outb %%al,%%dx\n\t"
	 "incw %%dx\n"
	 "owait:\t"
	 "inb %%dx,%%al\n\t"
	 "andb $0x80,%%al\n\t"
	 "cmpb %%al,%%ah\n\t"
	 "je owait\n\t"
	 "decw %%dx\n\t"
	 "xorb $0x80,%%ah\n\t"
	 "movb $8,%%al\n\t"
	 "outb %%al,%%dx\n\t"
	 "inc %%ebx\n\t"
	 "loop sloop\n\t"
	 "movb %%ah,ack"
	 : /* no outputs */
	 : "g" (buf), "g" (len)
	 : "ax", "ebx", "ecx", "dx");
# elif defined _MSC_VER
  unsigned *a = &ack; /* kludge to get around MS C's brain-dead behaviour */

  __asm {
    mov dx,baseaddr
    mov cx,len
    push ds
    mov ds,word ptr a+2
    mov ah,byte ptr ack

    lds bx,buf
sloop:
    mov al,[bx]
    xor al,0xb
    add dx,2
    out dx,al
    sub dx,2
    and al,0xf7
    out dx,al
    inc dx

owait:
    in al,dx
    and al,0x80
    cmp al,ah
    jne nowait
    in al,0x60
    dec al    /* esc pressed? */
    jne owait
    je abort

nowait:
    dec dx
    mov ah,al
    mov al,8
    out dx,al

    inc bx
    loop sloop

abort:
    mov ds,word ptr a+2
    mov byte ptr ack,ah
    pop ds
  }

  if (inb (0x60) == 1) /* ESC pressed during the transfer */
    exit (-1);
# else
  register const unsigned char* buffer = buf;

  while (len--) {
    /* invert bits 0, 1 and 3 */
    register unsigned char byte = *buffer++ ^ 0x0b;
    outb (byte, ctrladdr);
    outb (byte & ~8, baseaddr); /* drop the -FLAG to low */
    while (ack == (0x80 & inb(stataddr))); /* wait for BUSY to change */
    outb(8, baseaddr); /* raise -FLAG again */
    ack ^= 0x80; /* store the new state of -BUSY */
  }
# endif
  return 0;
}

/** Receive data
 * @param buf		the data to be received
 * @param len		length of the data in bytes
 * @return		zero on success, nonzero on failure
 */
int
prlink_read (void* buf, unsigned len)
{
# ifdef __GNUC__
  __asm ("movw baseaddr,%%dx\n\t"
	 "movl %1,%%ecx\n\t"
	 "movl %0,%%ebx\n\t"
	 "movb ack,%%ah\n"
	 "rloop:\t"
	 "xorb %%al,%%al\n\t"
	 "outb %%al,%%dx\n\t"
	 "incw %%dx\n"
	 "iwait1:\t"
	 "inb %%dx,%%al\n\t"
	 "andb $0x80,%%al\n\t"
	 "cmpb %%ah,%%al\n\t"
	 "je iwait1\n\t"
	 "decw %%dx\n\t"
	 "movb $8,%%al\n\t"
	 "outb %%al,%%dx\n\t"
	 "addw $2,%%dx\n\t"
	 "movb $4,%%al\n\t"
	 "outb %%al,%%dx\n\t"
	 "inb %%dx,%%al\n\t"
	 "andb $0xf,%%al\n\t"
	 "movb %%al,(%%ebx)\n\t"
	 "subw $2,%%dx\n\t"
	 "xorb %%al,%%al\n\t"
	 "outb %%al,%%dx\n\t"
	 "incw %%dx\n"
	 "iwait2:\t"
	 "inb %%dx,%%al\n\t"
	 "andb $0x80,%%al\n\t"
	 "cmpb %%ah,%%al\n\t"
	 "jne iwait2\n\t"
	 "decw %%dx\n\t"
	 "movb $8,%%al\n\t"
	 "outb %%al,%%dx\n\t"
	 "addw $2,%%dx\n\t"
	 "inb %%dx,%%al\n\t"
	 "subw $2,%%dx\n\t"
	 "salb $4,%%al\n\t"
	 "orb (%%ebx),%%al\n\t"
	 "xorb $0xbb,%%al\n\t"
	 "movb %%al,(%%ebx)\n\t"
	 "inc %%ebx\n\t"
	 "loop rloop"
	 : /* no outputs */
	 : "g" (buf), "g" (len)
	 : "ax", "ebx", "ecx", "dx");
# elif defined _MSC_VER
  unsigned *a = &ack; /* kludge to get around MS C's brain-dead behaviour */

  __asm {
    mov dx,baseaddr
    mov cx,len
    push ds
    mov ds,word ptr a+2
    mov ah,byte ptr ack

    lds bx,buf

rloop:
    xor al,al
    out dx,al
    inc dx

iwait1:
    in al,dx
    and al,0x80
    cmp al,ah
    jne niwait1
    in al,0x60  /* esc pressed? */
    dec al
    jnz iwait1
    jz abort

niwait1:
    dec dx
    mov al,8
    out dx,al
    add dx,2
    mov al,4
    out dx,al
    in al,dx
    and al,0xf
    mov [bx],al

    sub dx,2
    xor al,al
    out dx,al
    inc dx

iwait2:
    in al,dx
    and al,0x80
    cmp al,ah
    je niwait2
    in al,0x60  /* esc pressed? */
    dec al
    jnz iwait2
    jz abort

niwait2:
    dec dx
    mov al,8
    out dx,al

    add dx,2
    in al,dx
    sub dx,2
        
#  if defined _M_I286 || defined _M_I386
    sal al,4
#  else
    sal al,1
    sal al,1
    sal al,1
    sal al,1
#  endif
    or al,[bx]
    xor al,0xbb
    mov [bx],al

    inc bx
    loop rloop
abort:
    pop ds
  }

  if (inb (0x60) == 1) /* ESC pressed during the transfer */
    exit (-1);
# else
  register unsigned char data;
  register unsigned char* buffer = buf;

  while (len--) {
    outb (4, ctrladdr); /* set the bidirectional signals to input */
    outb (0, baseaddr); /* drop -FLAG */
    while (ack == (0x80 & inb (stataddr))); /* wait for -BUSY to change */
    outb (8, baseaddr); /* raise -FLAG */

    data = inb (ctrladdr) & 0x0f; /* read the low nybble */

    outb (0, baseaddr); /* drop -FLAG */
    while (ack != (0x80 & inb (stataddr))); /* wait for -BUSY to change */

    data |= inb (ctrladdr) << 4; /* read the high nybble */

    outb (8, baseaddr); /* raise -FLAG */

    *buffer++ = data ^ 0xbb;
  }
# endif
  return 0;
}
#endif /* COMM_PC */
