/**
 * @file x1541.c
 * Transfer routines for the Commodore serial bus to PC cable
 * designed by Leopoldo Ghielmetti
 * @author Marko Mkel (marko.makela@nic.funet.fi)
 */

/*
 * Copyright  1994-1996 Marko Mkel and Olaf Seibert
 * Copyright  2001 Marko Mkel
 * Original Linux and Commodore 64/128/Vic-20 version by Marko Mkel
 * Ported to the PET and the Amiga series by Olaf Seibert
 * Restructured by Marko Mkel
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef COMM_PC
# include "info.h"
# include "x1541.h"
# include "pcpar.h"

/* X1541 cable constants */
/** the serial bus clock line */
#define CLK 2
/** the serial bus data line */
#define DATA 8

/** Open the communication channel
 * @param dev		name of the communication device
 * @param hostinfo	(output) the computer model information
 * @return		zero on success, nonzero on failure
 */
int
x1541_init (const char* dev, struct hostinfo* hostinfo)
{
  unsigned char detectbuf[5];
  if (!open_port (dev))
    return 1;
  outb (0, ctrladdr);
  if (x1541_write ("", 1))
    return 1;
  x1541_sr ();
  if (x1541_read (detectbuf, sizeof detectbuf))
    return 2;
  x1541_rs ();
  hostinfo->host = detectbuf[0];
  hostinfo->driver = (unsigned) detectbuf[1] | (unsigned) detectbuf[2] << 8;
  hostinfo->basic = (unsigned) detectbuf[3] | (unsigned) detectbuf[4] << 8;
  return 0;
}

/** Close the communication channel */
void
x1541_close (void)
{
  outb (0, ctrladdr); /* release the serial lines */
}

/** Switch the data direction from receive to send */
void
x1541_rs (void)
{
  do {
    /* drop CLK and DATA */
    outb (CLK | DATA, ctrladdr);
    /* raise CLK and DATA */
    outb (0, ctrladdr);
  }
  while (!(inb (ctrladdr) & CLK)); /* repeat until CLK==low */

  /* drop CLK */
  outb (CLK, ctrladdr);

  /* wait for DATA==low */
  while (!(inb (ctrladdr) & DATA));

  /* raise CLK */
  outb (0, ctrladdr);

  /* wait for DATA==high */
  while (inb (ctrladdr) & DATA);
}

/** Switch the data direction from send to receive */
void
x1541_sr (void)
{
  /* wait for CLK==DATA==low */
  while ((~inb (ctrladdr)) & (CLK | DATA));

  do {
    /* drop CLK */
    outb (CLK, ctrladdr);
    /* raise CLK */
    outb (0, ctrladdr);
  } /* repeat until CLK==low && DATA==high */
  while ((inb (ctrladdr) & (CLK | DATA)) != CLK);

  /* drop DATA */
  outb (DATA, ctrladdr);

  /* wait for CLK==high */
  while (inb (ctrladdr) & CLK);

  /* raise DATA */
  outb (0, ctrladdr);
}

/** Send data
 * @param buf		the data to be sent
 * @param len		length of the data in bytes
 * @return		zero on success, nonzero on failure
 */
int
x1541_write (const void* buf, unsigned len)
{
# ifdef __GNUC__
  __asm ("movw ctrladdr,%%dx\n\t"
	 "movl %1,%%ecx\n\t"
	 "movl %0,%%ebx\n\t"
	 "xorb %%al,%%al\n"
	 "sloop:\t"
	 "movb (%%ebx),%%ah\n\t"
	 "push %%ecx\n\t"
	 "movl $8,%%ecx\n"
	 "bsloop:\t"
	 "shrb $1,%%ah\n\t"
	 "adcb $7,%%al\n\t"
	 "andb $10,%%al\n\t"
	 "outb %%al,%%dx\n"
	 "owait1:\t"
	 "inb %%dx,%%al\n\t"
	 "notb %%al\n\t"
	 "andb $10,%%al\n\t"
	 "jnz owait1\n\t"
	 "outb %%al,%%dx\n"
	 "owait2:\t"
	 "inb %%dx,%%al\n\t"
	 "andb $3,%%al\n\t"
	 "andb $10,%%al\n\t"
	 "jnz owait2\n\t"
	 "loop bsloop\n\t"
	 "pop %%ecx\n\t"
	 "inc %%ebx\n\t"
	 "loop sloop"
	 : /* no outputs */
	 : "g" (buf), "g" (len)
	 : "ax", "ebx", "ecx", "dx");
# elif defined _MSC_VER
  __asm {
    mov dx,ctrladdr
    mov cx,len
    xor al,al

    push ds
    lds bx,buf
sloop:
    mov ah,[bx]
    push cx
    mov cx,8
bloop:
    shr ah,1
    adc al,7	/* CLK==2, DATA==8 */
    and al,10
    out dx,al
owait1:
    in al,dx
    not al
    and al,CLK | DATA
    jz nowait1
    in al,0x60
    dec al      /* esc pressed? */
    jne owait1
    pop cx
    jmp abort
nowait1:
    out dx,al
owait2:
    in al,dx
    and al,CLK | DATA
    jz nowait2
    in al,0x60
    dec al      /* esc pressed? */
    jne owait2
    pop cx
    jmp abort
nowait2:
    loop bloop
    pop cx

    inc bx
    loop sloop

abort:
    pop ds
  }

  if (inb (0x60) == 1) /* ESC pressed during the transfer */
    exit (-1);
# else
  register const unsigned char* buffer = buf;

  while (len--) {
    register int i;
    register unsigned char byte = *buffer++;

    for (i = 8; i--; byte >>= 1) {
      /* drop CLK or DATA, depending on the bit value */
      outb ((byte & 1) ? DATA : CLK, ctrladdr);

      /* wait for CLK==DATA==low */
      while ((~inb (ctrladdr)) & (CLK | DATA));

      /* raise CLK and DATA */
      outb (0, ctrladdr);

      /* wait for CLK==DATA==high */
      while (inb (ctrladdr) & (CLK | DATA));
    }
  }
# endif
  return 0;
}

/** Receive data
 * @param buf		the data to be received
 * @param len		length of the data in bytes
 * @return		zero on success, nonzero on failure
 */
int
x1541_read (void* buf, unsigned len)
{
# ifdef __GNUC__
  __asm ("movw ctrladdr,%%dx\n\t"
	 "movl %1,%%ecx\n\t"
	 "movl %0,%%ebx\n"
	 "rloop:\t"
	 "push %%ecx\n\t"
	 "movl $8,%%ecx\n"
	 "brloop:\t"
	 "inb %%dx,%%al\n\t"
	 "testb $10,%%al\n\t"
	 "jz brloop\n\t"
	 "xorb $10,%%al\n\t"
	 "outb %%al,%%dx\n\t"
	 "rorb $1,%%al\n\t"
	 "rorw $1,%%ax\n"
	 "iwait2:\t"
	 "inb %%dx,%%al\n\t"
	 "notb %%al\n\t"
	 "andb $10,%%al\n\t"
	 "jz iwait2\n\t"
	 "xorb %%al,%%al\n\t"
	 "outb %%al,%%dx\n\t"
	 "loop brloop\n\t"
	 "pop %%ecx\n\t"
	 "movb %%ah,(%%ebx)\n\t"
	 "inc %%ebx\n\t"
	 "loop rloop"
	 : /* no outputs */
	 : "g" (buf), "g" (len)
	 : "ax", "ebx", "ecx", "dx");
# elif defined _MSC_VER
  __asm {
    mov dx,ctrladdr
    mov cx,len

    push ds
    lds bx,buf

rloop:
    push cx
    mov cx,8
brloop:
    in al,dx
    test al,CLK | DATA
    jnz niwait1
    in al,0x60  ; is the esc key pressed?
    dec al
    jnz brloop
    pop cx
    jmp abort
niwait1:
    xor al,CLK | DATA
    out dx,al
    ror al,1
    ror ax,1

iwait2:
    in al,dx
    not al
    and al,CLK | DATA
    jnz niwait2
    in al,0x60  ; is the esc key pressed?
    dec al
    jnz iwait2
    pop cx
    jmp abort
niwait2:
    xor al,al
    out dx,al
    loop brloop

    pop cx
    mov [bx],ah
    inc bx
    loop rloop
abort:
    pop ds
  }

  if (inb (0x60) == 1) /* ESC pressed during the transfer */
    exit (-1);
# else
  register unsigned char* buffer = buf;

  while (len--) {
    register int i;
    register unsigned char data = 0;
    for (i = 8; i--; ) {
      register unsigned char regval;
      /* wait for CLK==low or DATA==low */
      while (!((regval = inb (ctrladdr)) & (CLK | DATA)));

      /* send an acknowledgement */
      outb (regval ^ (CLK | DATA), ctrladdr);

      /* read the data */
      data >>= 1;
      if (regval & DATA) data |= 128;

      /* wait for acknowledgement (CLK==high or DATA==high) */
      while (!((~inb (ctrladdr)) & (CLK | DATA)));

      /* raise CLK and DATA */
      outb (0, ctrladdr);
    }

    *buffer++ = data;
  }
# endif
  return 0;
}
#endif /* COMM_PC */
